@@ -96,7 +96,7 @@ class KModes(BaseEstimator, ClusterMixin):
     """
 
     def __init__(self, n_clusters=8, max_iter=100, cat_dissim=matching_dissim,
-                 init='Cao', n_init=1, verbose=0, random_state=None, n_jobs=1):
+                 init='Cao', n_init=10, verbose=0, random_state=None, n_jobs=1):
 
         self.n_clusters = n_clusters
         self.max_iter = max_iter
@@ -113,16 +113,23 @@ class KModes(BaseEstimator, ClusterMixin):
                       "Setting n_init to 1.")
             self.n_init = 1
 
-    def fit(self, X, y=None, **kwargs):
+    def fit(self, X, y=None, sample_weight=None, **kwargs):
         """Compute k-modes clustering.
 
         Parameters
         ----------
         X : array-like, shape=[n_samples, n_features]
+
+        sample_weight : sequence, default: None
+        The weight that is assigned to each individual data point when
+        updating the centroids.
         """
         X = pandas_to_numpy(X)
 
         random_state = check_random_state(self.random_state)
+        _validate_sample_weight(sample_weight, n_samples=X.shape[0],
+                                n_clusters=self.n_clusters)
+
         self._enc_cluster_centroids, self._enc_map, self.labels_, self.cost_, \
         self.n_iter_, self.epoch_costs_ = k_modes(
             X,
@@ -134,6 +141,7 @@ class KModes(BaseEstimator, ClusterMixin):
             self.verbose,
             random_state,
             self.n_jobs,
+            sample_weight
         )
         return self
 
@@ -175,12 +183,11 @@ class KModes(BaseEstimator, ClusterMixin):
     def cluster_centroids_(self):
         if hasattr(self, '_enc_cluster_centroids'):
             return decode_centroids(self._enc_cluster_centroids, self._enc_map)
-        else:
-            raise AttributeError("'{}' object has no attribute 'cluster_centroids_' "
-                                 "because the model is not yet fitted.")
+        raise AttributeError("'{}' object has no attribute 'cluster_centroids_' "
+                             "because the model is not yet fitted.")
 
 
-def labels_cost(X, centroids, dissim, membship=None):
+def labels_cost(X, centroids, dissim, membship=None, sample_weight=None):
     """Calculate labels and cost function given a matrix of points and
     a list of centroids for the k-modes algorithm.
     """
@@ -191,15 +198,17 @@ def labels_cost(X, centroids, dissim, membship=None):
     cost = 0.
     labels = np.empty(n_points, dtype=np.uint16)
     for ipoint, curpoint in enumerate(X):
+        weight = sample_weight[ipoint] if sample_weight is not None else 1
         diss = dissim(centroids, curpoint, X=X, membship=membship)
         clust = np.argmin(diss)
         labels[ipoint] = clust
-        cost += diss[clust]
+        cost += diss[clust] * weight
 
     return labels, cost
 
 
-def k_modes(X, n_clusters, max_iter, dissim, init, n_init, verbose, random_state, n_jobs):
+def k_modes(X, n_clusters, max_iter, dissim, init, n_init, verbose, random_state, n_jobs,
+            sample_weight=None):
     """k-modes algorithm"""
     random_state = check_random_state(random_state)
     if sparse.issparse(X):
@@ -212,8 +221,8 @@ def k_modes(X, n_clusters, max_iter, dissim, init, n_init, verbose, random_state
     X, enc_map = encode_features(X)
 
     n_points, n_attrs = X.shape
-    assert n_clusters <= n_points, "Cannot have more clusters ({}) " \
-                                   "than data points ({}).".format(n_clusters, n_points)
+    assert n_clusters <= n_points, f"Cannot have more clusters ({n_clusters}) " \
+                                   f"than data points ({n_points})."
 
     # Are there more n_clusters than unique rows? Then set the unique
     # rows as initial values and skip iteration.
@@ -229,25 +238,27 @@ def k_modes(X, n_clusters, max_iter, dissim, init, n_init, verbose, random_state
     seeds = random_state.randint(np.iinfo(np.int32).max, size=n_init)
     if n_jobs == 1:
         for init_no in range(n_init):
-            results.append(_k_modes_single(X, n_clusters, n_points, n_attrs, max_iter,
-                                           dissim, init, init_no, verbose, seeds[init_no]))
+            results.append(_k_modes_single(
+                X, n_clusters, n_points, n_attrs, max_iter, dissim, init, init_no,
+                verbose, seeds[init_no], sample_weight
+            ))
     else:
         results = Parallel(n_jobs=n_jobs, verbose=0)(
             delayed(_k_modes_single)(X, n_clusters, n_points, n_attrs, max_iter,
-                                     dissim, init, init_no, verbose, seed)
+                                     dissim, init, init_no, verbose, seed, sample_weight)
             for init_no, seed in enumerate(seeds))
     all_centroids, all_labels, all_costs, all_n_iters, all_epoch_costs = zip(*results)
 
     best = np.argmin(all_costs)
     if n_init > 1 and verbose:
-        print("Best run was number {}".format(best + 1))
+        print(f"Best run was number {best + 1}")
 
     return all_centroids[best], enc_map, all_labels[best], \
         all_costs[best], all_n_iters[best], all_epoch_costs[best]
 
 
 def _k_modes_single(X, n_clusters, n_points, n_attrs, max_iter, dissim, init, init_no,
-                    verbose, random_state):
+                    verbose, random_state, sample_weight=None):
     random_state = check_random_state(random_state)
     # _____ INIT _____
     if verbose:
@@ -264,29 +275,30 @@ def _k_modes_single(X, n_clusters, n_points, n_attrs, max_iter, dissim, init, in
         if len(init.shape) == 1:
             init = np.atleast_2d(init).T
         assert init.shape[0] == n_clusters, \
-            "Wrong number of initial centroids in init ({}, should be {})." \
-            .format(init.shape[0], n_clusters)
+            f"Wrong number of initial centroids in init ({init.shape[0]}, " \
+            f"should be {n_clusters})."
         assert init.shape[1] == n_attrs, \
-            "Wrong number of attributes in init ({}, should be {})." \
-            .format(init.shape[1], n_attrs)
+            f"Wrong number of attributes in init ({init.shape[1]}, " \
+            f"should be {n_attrs})."
         centroids = np.asarray(init, dtype=np.uint16)
     else:
         raise NotImplementedError
 
     if verbose:
         print("Init: initializing clusters")
-    membship = np.zeros((n_clusters, n_points), dtype=np.uint8)
+    membship = np.zeros((n_clusters, n_points), dtype=np.bool_)
     # cl_attr_freq is a list of lists with dictionaries that contain the
     # frequencies of values per cluster and attribute.
     cl_attr_freq = [[defaultdict(int) for _ in range(n_attrs)]
                     for _ in range(n_clusters)]
     for ipoint, curpoint in enumerate(X):
+        weight = sample_weight[ipoint] if sample_weight is not None else 1
         # Initial assignment to clusters
         clust = np.argmin(dissim(centroids, curpoint, X=X, membship=membship))
         membship[clust, ipoint] = 1
         # Count attribute values per cluster.
         for iattr, curattr in enumerate(curpoint):
-            cl_attr_freq[clust][iattr][curattr] += 1
+            cl_attr_freq[clust][iattr][curattr] += weight
     # Perform an initial centroid update.
     for ik in range(n_clusters):
         for iattr in range(n_attrs):
@@ -303,35 +315,38 @@ def _k_modes_single(X, n_clusters, n_points, n_attrs, max_iter, dissim, init, in
     labels = None
     converged = False
 
-    _, cost = labels_cost(X, centroids, dissim, membship)
+    _, cost = labels_cost(X, centroids, dissim, membship, sample_weight)
 
     epoch_costs = [cost]
     while itr < max_iter and not converged:
         itr += 1
-        centroids, moves = _k_modes_iter(
+        centroids, cl_attr_freq, membship, moves = _k_modes_iter(
             X,
             centroids,
             cl_attr_freq,
             membship,
             dissim,
-            random_state
+            random_state,
+            sample_weight
         )
         # All points seen in this iteration
-        labels, ncost = labels_cost(X, centroids, dissim, membship)
+        labels, ncost = labels_cost(X, centroids, dissim, membship, sample_weight)
         converged = (moves == 0) or (ncost >= cost)
         epoch_costs.append(ncost)
         cost = ncost
         if verbose:
-            print("Run {}, iteration: {}/{}, moves: {}, cost: {}"
-                  .format(init_no + 1, itr, max_iter, moves, cost))
+            print(f"Run {init_no + 1}, iteration: {itr}/{max_iter}, "
+                  f"moves: {moves}, cost: {cost}")
 
     return centroids, labels, cost, itr, epoch_costs
 
 
-def _k_modes_iter(X, centroids, cl_attr_freq, membship, dissim, random_state):
+def _k_modes_iter(X, centroids, cl_attr_freq, membship, dissim, random_state,
+                  sample_weight):
     """Single iteration of k-modes clustering algorithm"""
     moves = 0
     for ipoint, curpoint in enumerate(X):
+        weight = sample_weight[ipoint] if sample_weight is not None else 1
         clust = np.argmin(dissim(centroids, curpoint, X=X, membship=membship))
         if membship[clust, ipoint]:
             # Point is already in its right place.
@@ -342,7 +357,8 @@ def _k_modes_iter(X, centroids, cl_attr_freq, membship, dissim, random_state):
         old_clust = np.argwhere(membship[:, ipoint])[0][0]
 
         cl_attr_freq, membship, centroids = _move_point_cat(
-            curpoint, ipoint, clust, old_clust, cl_attr_freq, membship, centroids
+            curpoint, ipoint, clust, old_clust, cl_attr_freq, membship, centroids,
+            weight
         )
 
         # In case of an empty cluster, reinitialize with a random point
@@ -353,14 +369,15 @@ def _k_modes_iter(X, centroids, cl_attr_freq, membship, dissim, random_state):
             rindx = random_state.choice(choices)
 
             cl_attr_freq, membship, centroids = _move_point_cat(
-                X[rindx], rindx, old_clust, from_clust, cl_attr_freq, membship, centroids
+                X[rindx], rindx, old_clust, from_clust, cl_attr_freq, membship,
+                centroids, weight
             )
 
-    return centroids, moves
+    return centroids, cl_attr_freq, membship, moves
 
 
 def _move_point_cat(point, ipoint, to_clust, from_clust, cl_attr_freq,
-                    membship, centroids):
+                    membship, centroids, sample_weight):
     """Move point between clusters, categorical attributes."""
     membship[to_clust, ipoint] = 1
     membship[from_clust, ipoint] = 0
@@ -370,7 +387,7 @@ def _move_point_cat(point, ipoint, to_clust, from_clust, cl_attr_freq,
         from_attr_counts = cl_attr_freq[from_clust][iattr]
 
         # Increment the attribute count for the new "to" cluster
-        to_attr_counts[curattr] += 1
+        to_attr_counts[curattr] += sample_weight
 
         current_attribute_value_freq = to_attr_counts[curattr]
         current_centroid_value = centroids[to_clust][iattr]
@@ -380,7 +397,7 @@ def _move_point_cat(point, ipoint, to_clust, from_clust, cl_attr_freq,
             centroids[to_clust][iattr] = curattr
 
         # Decrement the attribute count for the old "from" cluster
-        from_attr_counts[curattr] -= 1
+        from_attr_counts[curattr] -= sample_weight
 
         old_centroid_value = centroids[from_clust][iattr]
         if old_centroid_value == curattr:
@@ -389,3 +406,19 @@ def _move_point_cat(point, ipoint, to_clust, from_clust, cl_attr_freq,
             centroids[from_clust][iattr] = get_max_value_key(from_attr_counts)
 
     return cl_attr_freq, membship, centroids
+
+
+def _validate_sample_weight(sample_weight, n_samples, n_clusters):
+    if sample_weight is not None:
+        if len(sample_weight) != n_samples:
+            raise ValueError("sample_weight should be of equal size as samples.")
+        if any(
+                not isinstance(weight, int) and not isinstance(weight, float)
+                for weight in sample_weight
+        ):
+            raise ValueError("sample_weight elements should either be int or floats.")
+        if any(sample < 0 for sample in sample_weight):
+            raise ValueError("sample_weight elements should be positive.")
+        if sum([x > 0 for x in sample_weight]) < n_clusters:
+            raise ValueError("Number of non-zero sample_weight elements should be "
+                             "larger than the number of clusters.")
