@@ -2,7 +2,7 @@
 K-prototypes clustering for mixed categorical and numerical data
 """
 
-# pylint: disable=super-on-old-class,unused-argument,attribute-defined-outside-init
+# pylint: disable=unused-argument,attribute-defined-outside-init
 
 from collections import defaultdict
 
@@ -130,23 +130,31 @@ class KPrototypes(kmodes.KModes):
                       "Setting n_init to 1.")
             self.n_init = 1
 
-    def fit(self, X, y=None, categorical=None):
+    def fit(self, X, y=None, categorical=None, sample_weight=None):
         """Compute k-prototypes clustering.
 
         Parameters
         ----------
         X : array-like, shape=[n_samples, n_features]
         categorical : Index of columns that contain categorical data
+
+        sample_weight : sequence, default: None
+        The weight that is assigned to each individual data point when
+        updating the centroids.
+
         """
         if categorical is not None:
-            assert isinstance(categorical, (int, list, tuple)), "The 'categorical' \
+            assert isinstance(categorical, (int, list, tuple)), f"The 'categorical' \
                 argument needs to be an integer with the index of the categorical \
                 column in your data, or a list or tuple of several of them, \
-                but it is a {}.".format(type(categorical))
+                but it is a {type(categorical)}."
 
         X = pandas_to_numpy(X)
 
         random_state = check_random_state(self.random_state)
+        kmodes._validate_sample_weight(sample_weight, n_samples=X.shape[0],
+                                       n_clusters=self.n_clusters)
+
         # If self.gamma is None, gamma will be automatically determined from
         # the data. The function below returns its value.
         self._enc_cluster_centroids, self._enc_map, self.labels_, self.cost_, \
@@ -162,12 +170,13 @@ class KPrototypes(kmodes.KModes):
             self.n_init,
             self.verbose,
             random_state,
-            self.n_jobs
+            self.n_jobs,
+            sample_weight,
         )
 
         return self
 
-    def predict(self, X, categorical=None):
+    def predict(self, X, categorical=None, **kwargs):
         """Predict the closest cluster each sample in X belongs to.
 
         Parameters
@@ -184,10 +193,10 @@ class KPrototypes(kmodes.KModes):
         assert hasattr(self, '_enc_cluster_centroids'), "Model not yet fitted."
 
         if categorical is not None:
-            assert isinstance(categorical, (int, list, tuple)), "The 'categorical' \
+            assert isinstance(categorical, (int, list, tuple)), f"The 'categorical' \
                 argument needs to be an integer with the index of the categorical \
                 column in your data, or a list or tuple of several of them, \
-                but it is a {}.".format(type(categorical))
+                but it is a {type(categorical)}."
 
         X = pandas_to_numpy(X)
         Xnum, Xcat = _split_num_cat(X, categorical)
@@ -203,12 +212,12 @@ class KPrototypes(kmodes.KModes):
                 self._enc_cluster_centroids[0],
                 decode_centroids(self._enc_cluster_centroids[1], self._enc_map)
             ))
-        else:
-            raise AttributeError("'{}' object has no attribute 'cluster_centroids_' "
-                                 "because the model is not yet fitted.")
+        raise AttributeError("'{}' object has no attribute 'cluster_centroids_' "
+                             "because the model is not yet fitted.")
 
 
-def labels_cost(Xnum, Xcat, centroids, num_dissim, cat_dissim, gamma, membship=None):
+def labels_cost(Xnum, Xcat, centroids, num_dissim, cat_dissim, gamma,
+                membship=None, sample_weight=None):
     """Calculate labels and cost function given a matrix of points and
     a list of centroids for the k-prototypes algorithm.
     """
@@ -226,13 +235,16 @@ def labels_cost(Xnum, Xcat, centroids, num_dissim, cat_dissim, gamma, membship=N
         tot_costs = num_costs + gamma * cat_costs
         clust = np.argmin(tot_costs)
         labels[ipoint] = clust
-        cost += tot_costs[clust]
+        if sample_weight is not None:
+            cost += tot_costs[clust] * sample_weight[ipoint]
+        else:
+            cost += tot_costs[clust]
 
     return labels, cost
 
 
 def k_prototypes(X, categorical, n_clusters, max_iter, num_dissim, cat_dissim,
-                 gamma, init, n_init, verbose, random_state, n_jobs):
+                 gamma, init, n_init, verbose, random_state, n_jobs, sample_weight=None):
     """k-prototypes algorithm"""
     random_state = check_random_state(random_state)
     if sparse.issparse(X):
@@ -254,8 +266,8 @@ def k_prototypes(X, categorical, n_clusters, max_iter, num_dissim, cat_dissim,
     ncatattrs = len(categorical)
     nnumattrs = X.shape[1] - ncatattrs
     n_points = X.shape[0]
-    assert n_clusters <= n_points, "Cannot have more clusters ({}) " \
-                                   "than data points ({}).".format(n_clusters, n_points)
+    assert n_clusters <= n_points, f"Cannot have more clusters ({n_clusters}) " \
+                                   f"than data points ({n_points})."
 
     Xnum, Xcat = _split_num_cat(X, categorical)
     Xnum, Xcat = check_array(Xnum), check_array(Xcat, dtype=None)
@@ -287,19 +299,20 @@ def k_prototypes(X, categorical, n_clusters, max_iter, num_dissim, cat_dissim,
             results.append(_k_prototypes_single(Xnum, Xcat, nnumattrs, ncatattrs,
                                                 n_clusters, n_points, max_iter,
                                                 num_dissim, cat_dissim, gamma,
-                                                init, init_no, verbose, seeds[init_no]))
+                                                init, init_no, verbose, seeds[init_no],
+                                                sample_weight))
     else:
         results = Parallel(n_jobs=n_jobs, verbose=0)(
             delayed(_k_prototypes_single)(Xnum, Xcat, nnumattrs, ncatattrs,
                                           n_clusters, n_points, max_iter,
                                           num_dissim, cat_dissim, gamma,
-                                          init, init_no, verbose, seed)
+                                          init, init_no, verbose, seed, sample_weight)
             for init_no, seed in enumerate(seeds))
     all_centroids, all_labels, all_costs, all_n_iters, all_epoch_costs = zip(*results)
 
     best = np.argmin(all_costs)
     if n_init > 1 and verbose:
-        print("Best run was number {}".format(best + 1))
+        print(f"Best run was number {best + 1}")
 
     # Note: return gamma in case it was automatically determined.
     return all_centroids[best], enc_map, all_labels[best], all_costs[best], \
@@ -308,7 +321,7 @@ def k_prototypes(X, categorical, n_clusters, max_iter, num_dissim, cat_dissim,
 
 def _k_prototypes_single(Xnum, Xcat, nnumattrs, ncatattrs, n_clusters, n_points,
                          max_iter, num_dissim, cat_dissim, gamma, init, init_no,
-                         verbose, random_state):
+                         verbose, random_state, sample_weight=None):
     # For numerical part of initialization, we don't have a guarantee
     # that there is not an empty cluster, so we need to retry until
     # there is none.
@@ -333,16 +346,16 @@ def _k_prototypes_single(Xnum, Xcat, nnumattrs, ncatattrs, n_clusters, n_points,
                     for cur_init in init]
             assert init[0].shape[0] == n_clusters, \
                 "Wrong number of initial numerical centroids in init " \
-                "({}, should be {}).".format(init[0].shape[0], n_clusters)
+                f"({init[0].shape[0]}, should be {n_clusters})."
             assert init[0].shape[1] == nnumattrs, \
-                "Wrong number of numerical attributes in init ({}, should be {})." \
-                .format(init[0].shape[1], nnumattrs)
+                "Wrong number of numerical attributes in init " \
+                f"({init[0].shape[1]}, should be {nnumattrs})."
             assert init[1].shape[0] == n_clusters, \
-                "Wrong number of initial categorical centroids in init ({}, " \
-                "should be {}).".format(init[1].shape[0], n_clusters)
+                "Wrong number of initial categorical centroids in init " \
+                f"({init[1].shape[0]}, should be {n_clusters})."
             assert init[1].shape[1] == ncatattrs, \
-                "Wrong number of categorical attributes in init ({}, should be {})." \
-                .format(init[1].shape[1], ncatattrs)
+                "Wrong number of categorical attributes in init " \
+                f"({init[1].shape[1]}, should be {ncatattrs})."
             centroids = [np.asarray(init[0], dtype=np.float64),
                          np.asarray(init[1], dtype=np.uint16)]
         else:
@@ -360,29 +373,30 @@ def _k_prototypes_single(Xnum, Xcat, nnumattrs, ncatattrs, n_clusters, n_points,
 
         if verbose:
             print("Init: initializing clusters")
-        membship = np.zeros((n_clusters, n_points), dtype=np.uint8)
+        membship = np.zeros((n_clusters, n_points), dtype=np.bool_)
         # Keep track of the sum of attribute values per cluster so that we
         # can do k-means on the numerical attributes.
         cl_attr_sum = np.zeros((n_clusters, nnumattrs), dtype=np.float64)
         # Same for the membership sum per cluster
-        cl_memb_sum = np.zeros(n_clusters, dtype=int)
+        cl_memb_sum = np.zeros(n_clusters, dtype=np.float64)
         # cl_attr_freq is a list of lists with dictionaries that contain
         # the frequencies of values per cluster and attribute.
-        cl_attr_freq = [[defaultdict(int) for _ in range(ncatattrs)]
+        cl_attr_freq = [[defaultdict(float) for _ in range(ncatattrs)]
                         for _ in range(n_clusters)]
         for ipoint in range(n_points):
+            weight = sample_weight[ipoint] if sample_weight is not None else 1
             # Initial assignment to clusters
             clust = np.argmin(
                 num_dissim(centroids[0], Xnum[ipoint]) + gamma *
                 cat_dissim(centroids[1], Xcat[ipoint], X=Xcat, membship=membship)
             )
             membship[clust, ipoint] = 1
-            cl_memb_sum[clust] += 1
+            cl_memb_sum[clust] += weight
             # Count attribute values per cluster.
             for iattr, curattr in enumerate(Xnum[ipoint]):
-                cl_attr_sum[clust, iattr] += curattr
+                cl_attr_sum[clust, iattr] += curattr * weight
             for iattr, curattr in enumerate(Xcat[ipoint]):
-                cl_attr_freq[clust][iattr][curattr] += 1
+                cl_attr_freq[clust][iattr][curattr] += weight
 
         # If no empty clusters, then consider initialization finalized.
         if membship.sum(axis=1).min() > 0:
@@ -413,34 +427,35 @@ def _k_prototypes_single(Xnum, Xcat, nnumattrs, ncatattrs, n_clusters, n_points,
     converged = False
 
     _, cost = labels_cost(Xnum, Xcat, centroids,
-                          num_dissim, cat_dissim, gamma, membship)
+                          num_dissim, cat_dissim, gamma, membship, sample_weight)
 
     epoch_costs = [cost]
     while itr < max_iter and not converged:
         itr += 1
-        centroids, moves = _k_prototypes_iter(Xnum, Xcat, centroids,
-                                              cl_attr_sum, cl_memb_sum, cl_attr_freq,
-                                              membship, num_dissim, cat_dissim, gamma,
-                                              random_state)
+        centroids, cl_attr_sum, cl_memb_sum, cl_attr_freq, membship, moves = \
+            _k_prototypes_iter(Xnum, Xcat, centroids, cl_attr_sum, cl_memb_sum,
+                               cl_attr_freq, membship, num_dissim, cat_dissim,
+                               gamma, random_state, sample_weight)
 
         # All points seen in this iteration
         labels, ncost = labels_cost(Xnum, Xcat, centroids,
-                                    num_dissim, cat_dissim, gamma, membship)
+                                    num_dissim, cat_dissim, gamma, membship, sample_weight)
         converged = (moves == 0) or (ncost >= cost)
         epoch_costs.append(ncost)
         cost = ncost
         if verbose:
-            print("Run: {}, iteration: {}/{}, moves: {}, ncost: {}"
-                  .format(init_no + 1, itr, max_iter, moves, ncost))
+            print(f"Run: {init_no + 1}, iteration: {itr}/{max_iter}, "
+                  f"moves: {moves}, ncost: {ncost}")
 
     return centroids, labels, cost, itr, epoch_costs
 
 
 def _k_prototypes_iter(Xnum, Xcat, centroids, cl_attr_sum, cl_memb_sum, cl_attr_freq,
-                       membship, num_dissim, cat_dissim, gamma, random_state):
+                       membship, num_dissim, cat_dissim, gamma, random_state, sample_weight):
     """Single iteration of the k-prototypes algorithm"""
     moves = 0
     for ipoint in range(Xnum.shape[0]):
+        weight = sample_weight[ipoint] if sample_weight is not None else 1
         clust = np.argmin(
             num_dissim(centroids[0], Xnum[ipoint]) +
             gamma * cat_dissim(centroids[1], Xcat[ipoint], X=Xcat, membship=membship)
@@ -456,12 +471,12 @@ def _k_prototypes_iter(Xnum, Xcat, centroids, cl_attr_sum, cl_memb_sum, cl_attr_
         # Note that membship gets updated by kmodes.move_point_cat.
         # move_point_num only updates things specific to the k-means part.
         cl_attr_sum, cl_memb_sum = _move_point_num(
-            Xnum[ipoint], clust, old_clust, cl_attr_sum, cl_memb_sum
+            Xnum[ipoint], clust, old_clust, cl_attr_sum, cl_memb_sum, weight
         )
         # noinspection PyProtectedMember
         cl_attr_freq, membship, centroids[1] = kmodes._move_point_cat(
             Xcat[ipoint], ipoint, clust, old_clust,
-            cl_attr_freq, membship, centroids[1]
+            cl_attr_freq, membship, centroids[1], weight
         )
 
         # Update old and new centroids for numerical attributes using
@@ -481,22 +496,23 @@ def _k_prototypes_iter(Xnum, Xcat, centroids, cl_attr_sum, cl_memb_sum, cl_attr_
             rindx = random_state.choice(choices)
 
             cl_attr_sum, cl_memb_sum = _move_point_num(
-                Xnum[rindx], old_clust, from_clust, cl_attr_sum, cl_memb_sum
+                Xnum[rindx], old_clust, from_clust, cl_attr_sum, cl_memb_sum,
+                weight
             )
             cl_attr_freq, membship, centroids[1] = kmodes._move_point_cat(
                 Xcat[rindx], rindx, old_clust, from_clust,
-                cl_attr_freq, membship, centroids[1]
+                cl_attr_freq, membship, centroids[1], weight
             )
 
-    return centroids, moves
+    return centroids, cl_attr_sum, cl_memb_sum, cl_attr_freq, membship, moves
 
 
-def _move_point_num(point, to_clust, from_clust, cl_attr_sum, cl_memb_sum):
+def _move_point_num(point, to_clust, from_clust, cl_attr_sum, cl_memb_sum, sample_weight):
     """Move point between clusters, numerical attributes."""
     # Update sum of attributes in cluster.
     for iattr, curattr in enumerate(point):
-        cl_attr_sum[to_clust][iattr] += curattr
-        cl_attr_sum[from_clust][iattr] -= curattr
+        cl_attr_sum[to_clust][iattr] += curattr * sample_weight
+        cl_attr_sum[from_clust][iattr] -= curattr * sample_weight
     # Update sums of memberships in cluster
     cl_memb_sum[to_clust] += 1
     cl_memb_sum[from_clust] -= 1
